/*  fft.C
 *
 *  Copyright (C) 1999 Richard Boulton <richard@tartarus.org>
 *  Convolution stuff by Ralph Loader <suckfish@ihug.co.nz>
 *
 *  2000 Fabio Checconi <fchecconi@libero.it>: added fftw code
 *  2001 Fabio Checconi: rewritten in C++
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * The non-fftw class does no check on the number of elements, even if it works
 * only for powers of 2.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "fft.h"

#ifdef HAVE_FFTW

template <typename T>
void
Tuner::Fft<T>::init(int n) throw(std::bad_alloc)
{
	points = n;
	plan = rfftw_create_plan(points, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
	in = new fftw_real[points];
	out = new fftw_real[points];
}

template <typename T>
void
Tuner::Fft<T>::destroy() throw()
{
	delete []out;
	delete []in;
	rfftw_destroy_plan(plan);
}

template <typename T>
void
Tuner::Fft<T>::_applyWindow(unsigned char *ptr) throw()
{
	int i;
	if(!win)
		for(i = 0; i < points; ++i)
			*(in + i) = *(ptr + i);
	else
		for(i = 0; i < points; ++i)
			*(in + i) = *(ptr + i) * win[i];
}

template <>
void
Tuner::Fft<int>::_applyWindow(unsigned char *ptr) throw()
{
	int i;
	switch(order) {
		case LITTLEENDIAN:
			for(i = 0; i < points; ++i)
				*(in + i) = (ptr[2 * i] |
				(static_cast<int>(ptr[2 * i + 1]) << 8)) *
				win[i];
			break;
		case BIGENDIAN:
			for(i = 0; i < points; ++i)
				*(in + i) = (ptr[2 * i + 1] |
				(static_cast<int>(ptr[2 * i]) << 8)) *
				win[i];
	}
}

template <typename T>
void
Tuner::Fft<T>::applyWindow(T *ptr) throw()
{
	int i;
	if(!win)
		for(i = 0; i < points; ++i)
			*(in + i) = *(ptr + i);
	else
		for(i = 0; i < points; ++i)
			*(in + i) = *(ptr + i) * win[i];
}

template <typename T>
bool
Tuner::Fft<T>::_calc(unsigned char *ibuf, double *obuf) throw()
{
	_applyWindow(ibuf);
	rfftw_one(plan, in, out);
	obuf[0] = out[0] * out[0];
	for(int i = 1; i < (points + 1) / 2; ++i)
		obuf[i] = out[i]*out[i] + out[points - i]*out[points - i];
	if(points % 2 == 0)
		obuf[points / 2] = out[points / 2] *
			out[points / 2];

	return true;
}

template <typename T>
bool
Tuner::Fft<T>::calc(T *ibuf, double *obuf) throw()
{
	applyWindow(ibuf);
	rfftw_one(plan, in, out);
	obuf[0] = out[0] * out[0];
	for(int i = 1; i < (points + 1) / 2; ++i)
		obuf[i] = out[i]*out[i] + out[points - i]*out[points - i];
	if(points % 2 == 0)
		obuf[points / 2] = out[points / 2] *
			out[points / 2];

	return true;
}

#else

template <typename T>
int
Tuner::Fft<T>::reverseBits(int i) throw()
{
	int reversed = 0, loop;

	for(loop = 0; loop < lognn; ++loop) {
		reversed <<= 1;
		reversed += (i & 1);
		i >>= 1;
	}

	return reversed;
}

template <typename T>
int
Tuner::Fft<T>::log2(int n) throw()
{
	int ret = 0;
	while(n > 1)
		n >>= 1, ++ret;

	return ret;
}

template <typename T>
void
Tuner::Fft<T>::doTransform() throw()
{
	unsigned int i, j, k;
	unsigned int exchanges;
	double factReal, factImag;
	double tmpReal, tmpImag;
	unsigned int factfact;
    
	exchanges = 1;
	factfact = points / 2;

	for( i = lognn; i != 0; i-- ) {
		for( j = 0; j != exchanges; j++ ) {
			factReal = cosTable[j * factfact];
	 		factImag = sinTable[j * factfact];
	    
	  		for( k = j; k < points; k += exchanges << 1) {
				int k1 = k + exchanges;
				tmpReal = factReal * realOut[k1] - factImag *
					imagOut[k1];
				tmpImag = factReal * imagOut[k1] + factImag *
					realOut[k1];
				realOut[k1] = realOut[k] - tmpReal;
				imagOut[k1] = imagOut[k] - tmpImag;
				realOut[k]  += tmpReal;
				imagOut[k]  += tmpImag;
			}
		}
		exchanges <<= 1;
		factfact >>= 1;
	}
}

template <typename T>
void
Tuner::Fft<T>::_applyWindow(unsigned char *ptr) throw()
{
	int i;
	if(!win)
		for(i = 0; i < points; ++i)
			realOut[indexTable[i]] = ptr[i], imagOut[i] = 0;
	else
		for(i = 0; i < points; ++i)
			realOut[indexTable[i]] = ptr[i] * win[i],
				imagOut[i] = 0;
}

template <>
void
Tuner::Fft<int>::_applyWindow(unsigned char *ptr) throw()
{
	int i;
	switch(order) {
		case LITTLEENDIAN:
			for(i = 0; i < points; ++i)
				realOut[indexTable[i]] = (ptr[i * 2] |
				(static_cast<int>(ptr[i * 2 + 1]) << 8)) *
				win[i], imagOut[i] = 0;
			break;
		case BIGENDIAN:
			for(i = 0; i < points; ++i)
				realOut[indexTable[i]] = (ptr[i * 2 + 1] |
				(static_cast<int>(ptr[i * 2]) << 8)) *
				win[i], imagOut[i] = 0;
	}
}

template <typename T>
void
Tuner::Fft<T>::applyWindow(T *ptr) throw()
{
	int i;
	if(!win)
		for(i = 0; i < points; ++i)
			realOut[indexTable[i]] = ptr[i], imagOut[i] = 0;
	else
		for(i = 0; i < points; ++i)
			realOut[indexTable[i]] = ptr[i] * win[i],
				imagOut[i] = 0;
}


template <typename T>
void
Tuner::Fft<T>::init(int n) throw(std::bad_alloc)
{
	points = n;
	lognn = log2(n);

	int size = points >> 1;
	cosTable = new double[size];
	sinTable = new double[size];

	realOut = new double[points];
	imagOut = new double[points];

	indexTable = new int[n];

	int i;
	for(i = 0; i < points; ++i)
		indexTable[i] = reverseBits(i);

	for(i = 0; i < size; i++ ) {
		double j = 2 * PI * i / points;
		cosTable[i] = std::cos(j);
		sinTable[i] = std::sin(j);
   	}
}

template <typename T>
void
Tuner::Fft<T>::destroy() throw()
{
	delete []indexTable;
	delete []imagOut;
	delete []realOut;
	delete []sinTable;
	delete []cosTable;
}

template <typename T>
bool
Tuner::Fft<T>::_calc(unsigned char *ibuf, double *obuf) throw()
{
	_applyWindow(ibuf);
	doTransform();
	obuf[0] = realOut[0] * realOut[0];
	for(int i = 1; i < (points + 1) / 2; ++i)
		obuf[i] = realOut[i]*realOut[i] * imagOut[i]*imagOut[i];
	if(points % 2 == 0)
		obuf[points / 2] = realOut[points / 2] *
			realOut[points / 2];

	return true;

}

template <typename T>
bool
Tuner::Fft<T>::calc(T *ibuf, double *obuf) throw()
{
	applyWindow(ibuf);
	doTransform();
	obuf[0] = realOut[0] * realOut[0];
	for(int i = 1; i < (points + 1) / 2; ++i)
		obuf[i] = realOut[i]*realOut[i] * imagOut[i]*imagOut[i];
	if(points % 2 == 0)
		obuf[points / 2] = realOut[points / 2] *
			realOut[points / 2];

	return true;

}

#endif

template class Tuner::Fft<unsigned char>;
template class Tuner::Fft<int>;

