/*  GuiTuner 0.05
 *  Copyright (c) 2001 Fabio Checconi <fchecconi@libero.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * SEE EXCEPTIONS & SPEC
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cstring>
#include <cstdio>
#include <fcntl.h>
#include <unistd.h>
#include "genconf.h"
#include "i18n.h"

namespace Config {
	const char *EnvConfFileEntry = "GUITUNERRC";
	const char *DefaultRCName = "/.guitunerrc";
	const char *DefaultGlobalRC[2] = { "/usr/local/etc/guitunerrc",
		"/etc/guitunerrc" };

	const char *argList = "c:d:m:p:r:";
	const char *MainVal[3] = { "AudioDriver", "AudioDriverPlay", "DetectionMode" };
}

Config::GenConf::GenConf(int &argc, char **&argv, AudioDriver ad,
	AudioDriver adp, DetectionMode dm, AudioInfo *ai, AudioInfo *aip,
	CalcInfo *ci, NoteInfo *ni) throw(Error::ConfError)
{
	ConfFile cf;
	setInfo(ad, adp, dm, ai, aip, ci, ni);
	try {
		parseEnv();
		cf.load(inConf.c_str());
		load(cf);
	} catch(...) {
		loadDefaults();
	}
	parseCmdLine(argc, argv);
}

Config::GenConf::~GenConf()
{
	save();
}

bool
Config::GenConf::setInfo(AudioDriver ad, AudioDriver adp, DetectionMode d,
	const AudioInfo *ai, const AudioInfo *aip, const CalcInfo *fi,
	const NoteInfo *ni) throw()
{
	audr = ad, audrPlay = adp, dm = d;
	if(ai)
		audio = *ai;
	else
		audio.loadDefaults();
	if(aip)
		audioPlay = *aip;
	else
		audioPlay.loadDefaults();

	if(fi)
		fft = *fi;
	else
		fft.loadDefaults();

	if(ni)
		note = *ni;
	else
		note.loadDefaults();

	update();

	return true;
}

bool
Config::GenConf::load(Config::ConfFile &f) throw(Error::ConfError, Error::AudioError)
{
	std::string val[3];
	int v[3];

	for(int i = 0; i < 3; ++i) {
		f.getValue(MainVal[i], val[i]);
		v[i] = std::atoi(val[i].c_str());
	}

	if(val[0] != "" && v[0] >= AudioDriver(0) && v[0] <= Tuner::OSS_DRIVER)
		audr = AudioDriver(v[0]);
	if(val[1] != "" && v[1] >= AudioDriver(0) && v[1] <= Tuner::ESD_DRIVER)
		audrPlay = AudioDriver(v[1]);
	if(val[2] != "" && v[2] >= DetectionMode(0) && v[2] <= Tuner::FFTCALC)
		dm = DetectionMode(v[2]);

	audio.load(f);
	audioPlay.load(f);
	fft.load(f);
	note.load(f);
	update();

	return true;
}

bool
Config::GenConf::loadDefaults() throw()
{
	audr = audrPlay = Tuner::OSS_DRIVER;
	dm = Tuner::FFTCALC;
	audio.loadDefaults();
	audioPlay.loadDefaults();
	fft.loadDefaults();
	note.loadDefaults();
	update();

	return true;
}

void
Config::GenConf::save() throw()
{
	ConfFile cf;
	try {
		save(cf);
		cf.save(outConf.c_str());
	} catch(...) {
	}
}

bool
Config::GenConf::save(Config::ConfFile &f) throw(Error::ConfError, Error::AudioError)
{
	char buf[3][50];

	std::sprintf(buf[0], "%d", audr);
	std::sprintf(buf[1], "%d", audrPlay);
	std::sprintf(buf[2], "%d", dm);

	for(int i = 0; i < 3; ++i)
		f.setValue(MainVal[i], buf[i]);

	audio.save(f);
	fft.save(f);
	note.save(f);
	return true;
}

void
Config::GenConf::update() throw()
{
	note.delta = static_cast<double>(audio.rate) /
		static_cast<double>(fft.points);
}

bool
Config::GenConf::sync() throw()
{
	Tuner::Note::setNamingConv(note.conv);
	Tuner::Note::setRef(note.ref);
	Tuner::Note::setDelta(note.delta);

	return true;
}

void
Config::GenConf::parseCmdLine(int &argc, char **&argv) throw(Error::ConfError)
{
	char ch;
	opterr = 0;

	while((ch = getopt(argc, argv, argList)) != -1) {
		switch(ch) {
			case 'c':
				if(std::strcasecmp(optarg, "en") ||
					std::strcasecmp(optarg, "english"))
					note.conv = Tuner::ENGLISH;
				if(std::strcasecmp(optarg, "it") ||
					std::strcasecmp(optarg, "italian"))
					note.conv = Tuner::ITALIAN;
				break;
			case 'd':
				if(std::strcasecmp(optarg, "oss"))
					audr = Tuner::OSS_DRIVER;
				if(std::strcasecmp(optarg, "esd"))
					audr = Tuner::ESD_DRIVER;
				break;
			case 'm':
				if(std::strcasecmp(optarg, "fft"))
					dm = Tuner::FFTCALC;
				break;
			case 'p':
				if(int p = std::atoi(optarg) > 0)
					fft.points = p;
				break;
			case 'r':
				if(int p = std::atoi(optarg) > 0)
					audio.rate = p;
				break;

			default:;
		}
	}

	update();
}

void
Config::GenConf::parseEnv() throw(Error::ConfError)
{
	char *ptr = getenv(EnvConfFileEntry);
	if(ptr == 0) {
		ptr = getenv("HOME");
		inConf = outConf = std::string(ptr) + DefaultRCName;
	} else {
		inConf = outConf = ptr;
	}

	int fd = open(inConf.c_str(), O_RDONLY);
	if(fd == -1) {
		for(int i = 0; i < 2 && fd == -1; ++i)
			fd = open((inConf = DefaultGlobalRC[i]).c_str(), O_RDONLY);
	}

	if(fd == -1)
		throw Error::ConfError("Config::GenConf::parseEnv",
			_("Cannot find configuration file"));

	close(fd);
}

